--[[
LibPvpTalents-1.0 - Track the and retrieve info about PVP talents for all specializations.
Copyright (C) 2019-2020 Jordy141 (Kaelon-Nagrand)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Redistribution of a stand alone version is strictly prohibited without
      prior written authorization from the LibSpellbook project manager.
    * Neither the name of the LibSpellbook authors nor the names of its contributors
      may be used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--]]

local MAJOR, MINOR = "LibPvpTalents-1.0", 1
assert(LibStub, MAJOR.." requires LibStub")
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

-- constants
local _G = _G

-- blizzard api
local CreateFrame = _G.CreateFrame

-- lua api
local pairs = _G.pairs
local type = _G.type

lib.pvpTalents = {};
lib.availablePvpTalents = {};
lib.byName = {};
lib.byId = {};
lib.selectedTalents = {};
lib.pvpTalentSlotAmount = 0;

function lib:Init()
	lib.playerLevel = UnitLevel("player");
	local eventFrame = CreateFrame("FRAME", "LivPbpTalentsLevelUpFrame");
	eventFrame:RegisterEvent("PLAYER_PVP_TALENT_UPDATE");

	local maxLevel = (IsTrialAccount() or IsVeteranTrialAccount()) and 20 or GetAccountExpansionLevel() == 8 and 60 or 50;
	if(lib.playerLevel < maxLevel) then
		eventFrame:RegisterEvent("PLAYER_LEVEL_UP");
	end

	eventFrame:SetScript("OnEvent", function(self, event, ...)
		lib.playerLevel = UnitLevel("player");
		lib:RefreshTalents();
	end)

	lib:RefreshTalents();
end

function lib:CleanUp()
	lib.pvpTalents = {};
	lib.availablePvpTalents = {};
	lib.byName = {};
	lib.byId = {};
	lib.selectedTalents = {};
end

function lib:RefreshTalents()
	lib:CleanUp();
	lib.pvpTalentSlotAmount = 0;
	for i = 1, 3 do
		local talentSlotUnlockLevel = C_SpecializationInfo.GetPvpTalentSlotUnlockLevel(i);
		if(lib.playerLevel >= talentSlotUnlockLevel) then
			lib.pvpTalentSlotAmount = lib.pvpTalentSlotAmount + 1;
		end
	end
	
	for i = 1, 3 do
		local talentSlotInfo = C_SpecializationInfo.GetPvpTalentSlotInfo(i);
		if(talentSlotInfo.selectedTalentID) then
			table.insert(lib.selectedTalents, lib:GetPvpTalentInfoByID(talentSlotInfo.selectedTalentID))
		end

		if(#talentSlotInfo.availableTalentIDs > 0) then
			for key, id in pairs(talentSlotInfo.availableTalentIDs) do
				local pvpTalent = lib:GetPvpTalentInfoByID(id);
				if(not lib.byId[pvpTalent.talentID]) then
					lib.byId[pvpTalent.talentID] = pvpTalent;
					lib.byName[pvpTalent.name] = pvpTalent;
					table.insert(lib.pvpTalents, pvpTalent);
					if(lib.playerLevel >= C_SpecializationInfo.GetPvpTalentUnlockLevel(id)) then
						table.insert(lib.availablePvpTalents, pvpTalent);
					end
				end
			end
		end
	end
end

function lib:GetPvpTalentInfoByID(talentID)
	local talentID, name, texture, selected, available, spellID, unlocked, row, column, known, grantedByAura = GetPvpTalentInfoByID(talentID);
	return {
		talentID = talentID,
		name = name,
		texture = texture,
		selected = selected,
		available = available,
		spellID = spellID,
		unlocked = unlocked,
		row = row,
		column = column,
		known = known,

	};
end

function lib:IsTalentSelected(talentID)
	if talentID == nil then
		return nil;
	end

	for i, talent in pairs(self.selectedTalents) do 
		if talent.talentID == talentID then
			return true;
		end
	end 
	return false;
end

function lib:IsTalentSelectedBySpellID(spellID)
	if spellID == nil then
		return nil;
	end

	for i, talent in pairs(self.selectedTalents) do 
		if talent.spellID == spellID then
			return true;
		end
	end 
	return false;
end

function lib:IsTalentSelectedByName(name) 
	if name == nil then
		return nil;
	end

	for i, talent in pairs(self.selectedTalents) do 
		if talent.name == name then
			return true;
		end
	end 
	return false;
end

function lib:GetAllTalents() 
	return lib.pvpTalents;
end

function lib:GetAvailableTalents()
	return lib.availablePvpTalents;
end

function lib:GetTalentByID(talentID)
	if talentID == nil then
		return nil;
	end

	for i, talent in pairs(self.pvpTalents) do 
		if talent.talentID == talentID then
			return talent;
		end
	end 

	return nil;
end

function lib:GetTalentBySpellID(spellID) 
	if spellID == nil then
		return nil;
	end

	for i, talent in pairs(self.pvpTalents) do 
		if talent.spellID == spellID then
			return talent;
		end
	end 

	return nil;
end

function lib:GetTalentByName(name)
	if name == nil then
		return nil;
	end

	for i, talent in pairs(self.pvpTalents) do 
		if talent.name == name then
			return talent;
		end
	end 

	return nil;
end

function lib:GetAvailableTalentByID(talentID)
	if talentID == nil then
		return nil;
	end

	for i, talent in pairs(self.availablePvpTalents) do 
		if talent.talentID == talentID then
			return talent;
		end
	end 

	return nil;
end

function lib:GetAvailableTalentBySpellID(spellID) 
	if spellID == nil then
		return nil;
	end

	for i, talent in pairs(self.availablePvpTalents) do 
		if talent.spellID == spellID then
			return talent;
		end
	end 

	return nil;
end

function lib:GetAvailableTalentByName(name)
	if name == nil then
		return nil;
	end

	for i, talent in pairs(self.availablePvpTalents) do 
		if talent.name == name then
			return talent;
		end
	end 

	return nil;
end

function lib:TalentCanBeSelected(talent)
	if talent == nil then
		return false;
	end
	if(type(talent) == "table") then
		if(not talent.talentID) then
			return false;
		end
		talent = talent.talentID;
	end
	if(type(talent) == "number") then
		for i, pvpTalent in pairs(self.availablePvpTalents) do
			if pvpTalent.talentID == talent then
				return true;
			end
		end
	elseif(type(talent) == "string") then
		for i, pvpTalent in pairs(self.availablePvpTalents) do
			if pvpTalent.name == talent then
				return true;
			end
		end
	end
	return false;
end

function lib:TalentCanBeSelectedBySpellID(spellID)
	if type(spellID) ~= "number" then
		return false;
	end

	local talent = lib:GetAvailableTalentBySpellID(spellID);
	if talent == nil then
		return false
	end
	return lib:TalentCanBeSelected(talent);
end

function lib:GetAvailableTalentSlotAmount()
	return self.pvpTalentSlotAmount;
end

function lib:SlotIsAvailable(slotID)
	return self.pvpTalentSlotAmount >= slotID;
end

function lib:SlotIsEmpty(slotID)
	if slotID > self.pvpTalentSlotAmount then
		return false
	end
	
	return not C_SpecializationInfo.GetPvpTalentSlotInfo(slotID).selectedTalentID;
end

function lib:GetTalentBySlotID(slotID)
	local slotInfo = C_SpecializationInfo.GetPvpTalentSlotInfo(slotID);
	if not slotInfo.selectedTalentID then
		return nil;
	end
	return lib:GetTalentByID(slotInfo.selectedTalentID);
end

function lib:SetPvpTalent(talent, slotID)
	if type(slotID) ~= "number" or slotID > self.pvpTalentSlotAmount then
		return false
	end
	local talentID = nil;
	if type(talent) == "number" then
		talentID = talent
	elseif type(talent) == "table" then
		if not talent.talentID then
			return false
		end
		talentID = talent.talentID
	elseif (type(talent) == "string") then
		local pvpTalent = lib:GetAvailableTalentByName(talent);
		if pvpTalent == nil then
			return false
		end
		talentID = pvpTalent.talentID;
	end
	
	if talentID == nil then
		return false;
	end
	LearnPvpTalent(talentID, slotID);
	return true;
end

function lib:SetPvpTalentBySpellID(spellID, slotID)
	if type(spellID) ~= "number" then
		return false
	end
	local pvpTalent = lib:GetAvailableTalentBySpellID(spellID);
	if pvpTalent == nil then
		return false
	end
	return lib:SetPvpTalent(pvpTalent, slotID);
end

local loadFrame = CreateFrame("Frame", "LibPvpTalentsLoadFrame")
loadFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
loadFrame:SetScript("OnEvent", function(self, event, ...) lib:Init() end)